import numpy as np
import pandas as pd
import scipy.sparse as sp

import matplotlib.pyplot as plt
from matplotlib.patches import Arc, Polygon, Circle
import matplotlib.animation as animation
import seaborn as sns
import re
import h5py 


### Extruder funcs

class leg(object):
    def __init__(self, pos, attrs={"stalled":False, "CTCF":False}):
        """
        A leg has two important attribues: pos (positions) and attrs (a custom list of attributes)
        """
        self.pos = pos
        self.attrs = dict(attrs)
        
class cohesin(object):
    """
    A cohesin class provides fast access to attributes and positions 
    
    
    cohesin.left is a left leg of cohesin, cohesin.right is a right leg
    cohesin[-1] is also a left leg and cohesin[1] is a right leg         
    
    Also, cohesin.any("myattr") is True if myattr==True in at least one leg
    cohesin.all("myattr") is if myattr=True in both legs
    """
    def __init__(self, leg1, leg2):
        self.left = leg1
        self.right = leg2
   
    def any(self, attr):
        return self.left.attrs[attr] or self.right.attrs[attr]
    
    def all(self, attr):
        return self.left.attrs[attr] and self.right.attrs[attr]    
    
    def __getitem__(self, item):
        if item == -1:
            return self.left
        elif item == 1:
            return self.right 
        else:
            raise ValueError()

class RNAp(object):
    def __init__(self, position, direction, gene_end, elong_vel = 100, introns = []):
        self.position = position
        assert direction == 1 or direction == -1, f'direction ({direction}) is not eq 1 or -1'
        self.direction = direction
        self.gene_end = gene_end
        self.bristle = 0
        self.elong_vel = elong_vel # bp/s
        self.d_brist = 0
        self.state = 'elongation'
        self.introns = introns
        self.current_intron = 0
        
    def __str__(self):
        return f'[\
Position: {self.position},\n\
Direction: {self.direction},\n\
Gene_end: {self.gene_end},\n\
Bristle: {self.bristle},\n\
Elongation velocity: {self.elong_vel},\n\
D_brist: {self.d_brist},\n\
State: {self.state}]\
        '
    def __repr__(self):
        return self.__str__()
        
    def _elongation(self, args):
        self.bristle += self.elong_vel * args['TIMESTEP']
        self.d_brist += self.elong_vel * args['TIMESTEP']
        if len(self.introns) > self.current_intron:
            if self.introns[self.current_intron][0] <= self.bristle:
                self.bristle -= self.introns[self.current_intron][1]
                self.current_intron += 1
        
    def time_step(self, args, **kwargs):
        if self.state == 'elongation':
            self._elongation(args)
            if self.d_brist >= args['BEADSIZE']:
                self.d_brist -= args['BEADSIZE']
                if self.position == self.gene_end:
                    self.state = 'unload'
                else:
                    self.state = 'step'

        return self.do_step(args, **kwargs)
    
    def do_step(self, args, occupied_rnap):
        if self.state == 'step' and occupied_rnap[self.position + self.direction][(self.direction-1)//2] == 0:
            occupied_rnap[self.position][(self.direction-1)//2] = 0
            self.position += self.direction
            occupied_rnap[self.position][(self.direction-1)//2] = 1
            self.state = 'elongation'
            return True
        else:
            return False
            
def unloadProb(cohesin, args):
    """
    Defines unload probability based on a state of cohesin 
    """
    #Попробуем некоторые модификации вероятностей открепления для когезинов
    #
    
    if cohesin.any("stalled"):
        # if one side is stalled, we have different unloading probability 
        # Note that here we define stalled cohesins as those stalled not at CTCFs 
        return 1 / args["LIFETIME_STALLED"]
    # otherwise we are just simply unloading 
    return 1 / args["LIFETIME"]    
    
def loadProb(args):
    """
    Определяем вероятность присоединения когезина, основываясь на числе свободным мест
    """
    #return 1/(args["LIFETIME"]*args["LIFETIME"]/200-args["LIFETIME"])
    return 1 / args["BOUNDTIME"]


def loadOne_sisters(cohesins, occupied, args): 
    """
    A function to load one cohesin 
    """
    while True:
        a = np.random.randint(args["N"]//2)
        if (occupied[a] == 0) and (occupied[args["N"]-1-a] == 0):
            # Cohesin has mark "2" as occupied
            occupied[a] = 2
            occupied[args["N"]-1-a] = 2
            cohesins.append(cohesin(leg(a), leg(args["N"]-1-a)))
            break

def loadOne(condensins, occupied, args, n_trials = None, **kwargs): 
    """
    A function to load one cohesin 
    """
    n_trials = n_trials or 1000
    for _ in range(n_trials):
        a = np.random.randint(args["N"])
        if (occupied[a] == 0) and (occupied[a+1] == 0):
            if 'occupied_rnap' in kwargs.keys():
                occupied_rnap = kwargs['occupied_rnap']
                if (sum(occupied_rnap[a]) == 0) and (sum(occupied_rnap[a+1]) == 0):
                    pass
                else:
                    return False
            occupied[a] = -1
            occupied[a+1] = 1 
            condensins.append(cohesin(leg(a), leg(a+1)))
            break

def translocate_cohesin(cohesins, occupied, args, **kwargs):
    """
    This function describes everything that happens with cohesins - 
    loading/unloading them and stalling against each other 
    It relies on the functions defined above: unload probability, capture/release. 
    """
    # finally we translocate, and mark stalled cohesins because 
    # the unloadProb needs this 
    for i in range(len(cohesins)):
        cohesin = cohesins[i]
        dif_prob = args["Diffusion_prob"]
        for leg in [-1,1]:
            if np.random.random() < dif_prob:
                step_dif = 1-np.random.randint(2)*2
                if (occupied[cohesin[leg].pos + step_dif] == 0):
                    if 'occupied_rnap' in kwargs.keys():
                        occupied_rnap = kwargs['occupied_rnap']
                        if sum(occupied_rnap[cohesin[leg].pos + step_dif]) == 0:
                            cohesin[leg].attrs["stalled"] = False
                            occupied[cohesin[leg].pos] = 0
                            occupied[cohesin[leg].pos + step_dif] = 2
                            cohesin[leg].pos += step_dif
                    else:
                        cohesin[leg].attrs["stalled"] = False
                        occupied[cohesin[leg].pos] = 0
                        occupied[cohesin[leg].pos + step_dif] = 2
                        cohesin[leg].pos += step_dif

            cohesins[i] = cohesin

def push_cohesin_leg(cohesins, i, occupied, direction, args, chrom, **kwargs): # chrom - какая хроматида/какая нога когезина
    # Каким образом когезины сдвигаются под действием конденсинов and RNAp
    if 'occupied_rnap' in kwargs.keys():
        occupied_rnap = kwargs['occupied_rnap']
        if (sum(occupied_rnap[cohesins[i][chrom].pos + direction]) == 0):
            pass
        else:
            cohesins[i][chrom].attrs["stalled"] = True
            return False
    
    if (occupied[cohesins[i][chrom].pos + direction] == 0):
        cohesins[i][chrom].attrs["stalled"] = False 
        occupied[cohesins[i][chrom].pos] = 0
        occupied[cohesins[i][chrom].pos + direction] = 2
        cohesins[i][chrom].pos += direction
        return True
    elif (occupied[cohesins[i][chrom].pos + direction] == 2):
        for k in range(len(cohesins)):           
            if ((cohesins[i][chrom].pos + direction) == cohesins[k][-1].pos):
                if not push_cohesin_leg(cohesins, k, occupied, direction, args, -1, **kwargs): #Сдвигать все когезин
                    cohesins[i][chrom].attrs["stalled"] = True
                    return False
                else:
                    cohesins[i][chrom].attrs["stalled"] = False 
                    occupied[cohesins[i][chrom].pos] = 0
                    occupied[cohesins[i][chrom].pos + direction] = 2
                    cohesins[i][chrom].pos += direction
                    return True
                break
            elif ((cohesins[i][chrom].pos + direction) == cohesins[k][1].pos):
                if not push_cohesin_leg(cohesins, k, occupied, direction, args, 1, **kwargs): #Сдвигать все когезины
                    cohesins[i][chrom].attrs["stalled"] = True
                    return False
                else:
                    cohesins[i][chrom].attrs["stalled"] = False 
                    occupied[cohesins[i][chrom].pos] = 0
                    occupied[cohesins[i][chrom].pos + direction] = 2
                    cohesins[i][chrom].pos += direction
                    return True
                break
    else:
        cohesins[i][chrom].attrs["stalled"] = True
        return False
    
def push_condensin_leg(condensins, i, occupied, direction, args, leg, **kwargs):
    if 'occupied_rnap' in kwargs.keys():
        occupied_rnap = kwargs['occupied_rnap']
        if (sum(occupied_rnap[condensins[i][leg].pos + direction]) == 0):
            pass
        else:
            condensins[i][leg].attrs["stalled"] = True
            return False
    
    if (occupied[condensins[i][leg].pos + direction] == 0):
        condensins[i][leg].attrs["stalled"] = False 
        occupied[condensins[i][leg].pos] = 0
        occupied[condensins[i][leg].pos + direction] = leg
        condensins[i][leg].pos += direction
        return True
    elif (occupied[condensins[i][leg].pos + direction] == 2) and ('cohesins' in kwargs.keys()):
        cohesins = kwargs['cohesins']
        new_kwargs = kwargs.copy()
        _ = new_kwargs.pop('cohesins')
        for k in range(len(cohesins)):           
            if ((condensins[i][leg].pos + direction) == cohesins[k][-1].pos):
                if not push_cohesin_leg(cohesins, k, occupied, direction, args, -1, **new_kwargs): #Сдвигать все когезин
                    condensins[i][leg].attrs["stalled"] = True
                    return False
                else:
                    condensins[i][leg].attrs["stalled"] = False 
                    occupied[condensins[i][leg].pos] = 0
                    occupied[condensins[i][leg].pos + direction] = leg
                    condensins[i][leg].pos += direction
                    return True
                break
            elif ((condensins[i][leg].pos + direction) == cohesins[k][1].pos):
                if not push_cohesin_leg(cohesins, k, occupied, direction, args, 1, **new_kwargs): #Сдвигать все когезины
                    condensins[i][leg].attrs["stalled"] = True
                    return False
                else:
                    condensins[i][leg].attrs["stalled"] = False 
                    occupied[condensins[i][leg].pos] = 0
                    occupied[condensins[i][leg].pos + direction] = leg
                    condensins[i][leg].pos += direction
                    return True
                break
    elif abs(occupied[condensins[i][leg].pos + direction]) == 1:
        for k in range(len(condensins)):           
            if ((condensins[i][leg].pos + direction) == condensins[k][-1].pos):
                if not push_condensin_leg(condensins, k, occupied, direction, args, -1, **kwargs): #Сдвигать все когезин
                    condensins[i][leg].attrs["stalled"] = True
                    return False
                else:
                    condensins[i][leg].attrs["stalled"] = False 
                    occupied[condensins[i][leg].pos] = 0
                    occupied[condensins[i][leg].pos + direction] = leg
                    condensins[i][leg].pos += direction
                    return True
                break
            elif ((condensins[i][leg].pos + direction) == condensins[k][1].pos):
                if not push_condensin_leg(condensins, k, occupied, direction, args, 1, **kwargs): #Сдвигать все когезины
                    condensins[i][leg].attrs["stalled"] = True
                    return False
                else:
                    condensins[i][leg].attrs["stalled"] = False 
                    occupied[condensins[i][leg].pos] = 0
                    occupied[condensins[i][leg].pos + direction] = leg
                    condensins[i][leg].pos += direction
                    return True
                break
    else:
        condensins[i][leg].attrs["stalled"] = True
        return False

    
def cond_step_over_RNAp(pos, leg, args, **kwargs):
    if 'occupied_rnap' in kwargs.keys():
        occupied_rnap = kwargs['occupied_rnap']
        if (sum(occupied_rnap[pos + leg]) == 2) or (sum(occupied_rnap[pos + leg]) == 1 and args['COND_STEP_OVER_RNAP_PROB'] < np.random.random()):
            return False
        else:
            return True
    else:
        return True
        
def translocate_condensin(condensins, occupied, args, maxcondesins=500, **kwargs):
    """
    This function describes everything that happens with cohesins - 
    loading/unloading them and stalling against each other 
    
    It relies on the functions defined above: unload probability, capture/release. 
    """
    # first we try to unload cohesins and free the matching occupied sites
    linked = 0
    for _ in range(len(condensins)):
        prob = unloadProb(condensins[linked], args)
        # Сперва проверяем время ли конденсину отцепиться, если это так, то удаляем
        if np.random.random() < prob: #prob~1/lifetime or lifetime_stalled
            occupied[condensins[linked].left.pos] = 0 
            occupied[condensins[linked].right.pos] = 0
            del condensins[linked]
        elif 'occupied_rnap' in kwargs.keys():
            occupied_rnap = kwargs['occupied_rnap']
            leg_was_removed = False
            for leg in [-1,1]:
                if (sum(occupied_rnap[condensins[linked][leg].pos]) == 2) or (sum(occupied_rnap[condensins[linked][leg].pos]) == 1 and args['RNAP_REM_COND_PROB'] > np.random.random()):
                    occupied[condensins[linked].left.pos] = 0 
                    occupied[condensins[linked].right.pos] = 0
                    del condensins[linked]
                    leg_was_removed = True
                    break
            if not leg_was_removed:
                linked+=1
        else:
            linked+=1
        if linked == len(condensins) - 1: break
    
    for _ in range(int(maxcondesins-len(condensins))):
        prob = loadProb(args)
        # Сажаем новые конденсины с некоторой вероятностью
        if (np.random.random() < prob) & (len(condensins)<maxcondesins):
            loadOne(condensins, occupied, args, **kwargs) 
    
    for i in range(len(condensins)):
        condensin = condensins[i] 
        leg = int(np.random.random() < 0.5)*2-1
        # Normal condensin step block
        if occupied[condensin[leg].pos  + leg] == leg:
            condensin[leg].attrs["stalled"] = True
        elif (occupied[condensin[leg].pos  + leg] == -leg) or (occupied[condensin[leg].pos  + leg]>8):
            condensin[leg].attrs["stalled"] = True
        elif occupied[condensin[leg].pos  + leg] == 2:
            if 'cohesins' in kwargs.keys():
                cohesins = kwargs['cohesins']
                new_kwargs = kwargs.copy()
                _ = new_kwargs.pop('cohesins')
                for k in range(len(cohesins)):           
                    if ((condensin[leg].pos  + leg)==cohesins[k][-1].pos):
                        if not push_cohesin_leg(cohesins, k, occupied, leg, args, -1, **new_kwargs): #Сдвигать все когезин
                            condensin[leg].attrs["stalled"] = True
                        break
                    elif ((condensin[leg].pos  + leg)==cohesins[k][1].pos):
                        if not push_cohesin_leg(cohesins, k, occupied, leg, args, 1, **new_kwargs): #Сдвигать все когезины
                            condensin[leg].attrs["stalled"] = True
                        break
            if occupied[condensin[leg].pos  + leg] == 0 and cond_step_over_RNAp(condensin[leg].pos, leg, args, **kwargs):
                condensin[leg].attrs["stalled"] = False 
                occupied[condensin[leg].pos] = 0
                occupied[condensin[leg].pos + leg] = leg
                condensin[leg].pos += leg
            else:
                condensin[leg].attrs["stalled"] = True                
        elif cond_step_over_RNAp(condensin[leg].pos, leg, args, **kwargs):
            condensin[leg].attrs["stalled"] = False 
            occupied[condensin[leg].pos] = 0
            occupied[condensin[leg].pos + leg] = leg
            condensin[leg].pos += leg
        
        condensins[i] = condensin

    # Condensins legs too close check
    linked = 0
    for _ in range(len(condensins)):
        if abs(condensins[linked].left.pos - condensins[linked].right.pos) <= 1:
            occupied[condensins[linked].left.pos] = 0
            occupied[condensins[linked].right.pos] = 0
            del condensins[linked]
        else:
            linked+=1
            
def translocate_rnap(rnaps, occupied_rnap, chosen_genes_beads, args, **kwargs):
    # Unload part
    i_rnap = 0
    for _ in range(len(rnaps)):
        rnap = rnaps[i_rnap]
        if rnap.state != 'unload':
            i_rnap += 1
        else:
            occupied_rnap[rnap.position][(rnap.direction-1)//2] = 0
            del rnaps[i_rnap]
    
    # Load part
    for (st, dr, en, prob, introns) in chosen_genes_beads:
        if np.random.random() < prob and occupied_rnap[st][(dr-1)//2] == 0:
            rnaps.append(RNAp(st, dr, en, elong_vel = args['RNAP_VEL'], introns = introns))
            occupied_rnap[st][(dr-1)//2] = 1

    # Moving part
    for rnap in rnaps:
        stepped = rnap.time_step(args, occupied_rnap = occupied_rnap)
        if stepped and ('occupied' in kwargs.keys()):
            occupied = kwargs['occupied']
            for pos_shift in [0, -rnap.direction]:
                if (occupied[rnap.position + pos_shift] == 2) and ('cohesins' in kwargs.keys()):
                    cohesins = kwargs['cohesins']
                    for k in range(len(cohesins)):
                        if ((rnap.position + pos_shift) == cohesins[k][-1].pos):
                            push_cohesin_leg(cohesins, k, occupied, rnap.direction, args, -1)
                            break
                        elif ((rnap.position + pos_shift) == cohesins[k][1].pos):
                            push_cohesin_leg(cohesins, k, occupied, rnap.direction, args, 1)
                            break
                elif (abs(occupied[rnap.position + pos_shift]) == 1) and ('condensins' in kwargs.keys()):
                    condensins = kwargs['condensins']
                    for k in range(len(condensins)):
                        if ((rnap.position + pos_shift) == condensins[k][-1].pos):
                            new_kwargs = kwargs.copy()
                            _ = new_kwargs.pop('occupied')
                            _ = new_kwargs.pop('condensins')
                            if (not push_condensin_leg(condensins, k, occupied, rnap.direction, args, -1, **new_kwargs)) or (args['RNAP_REM_COND_PROB'] > np.random.random()):
                                occupied[condensins[k][-1].pos] = 0
                                occupied[condensins[k][1].pos] = 0
                                del(condensins[k])
                            break
                        elif ((rnap.position + pos_shift) == condensins[k][1].pos):
                            new_kwargs = kwargs.copy()
                            _ = new_kwargs.pop('occupied')
                            _ = new_kwargs.pop('condensins')
                            if (not push_condensin_leg(condensins, k, occupied, rnap.direction, args, 1, **new_kwargs)) or (args['RNAP_REM_COND_PROB'] > np.random.random()):
                                occupied[condensins[k][-1].pos] = 0
                                occupied[condensins[k][1].pos] = 0
                                del(condensins[k])
                            break

                            
                            
                            
### Simulations

def slice_str_parser(chosen_slice_str):
    chosen_slice = chosen_slice_str.split(':')
    chosen_slice = [chosen_slice[0], *chosen_slice[1].split('-')]
    chosen_slice[1] = int(chosen_slice[1].replace(',', ''))
    chosen_slice[2] = int(chosen_slice[2].replace(',', ''))
    return chosen_slice

def genes_to_beads_converter(chosen_genes, full_df, args):
    chosen_genes_beads = []
    for i in chosen_genes.index:
        if chosen_genes.loc[i, 'rnap_sep'] is not np.nan:
            direction = int(chosen_genes.loc[i, 'direction']+'1')

            start_pos = chosen_genes.loc[i, 'start'] if direction > 0 else chosen_genes.loc[i, 'end']
            start_pos = start_pos//args['BEADSIZE'] + direction*((start_pos%args['BEADSIZE']) < (args['BEADSIZE']//2))
            start_pos -= chosen_slice[1]//args['BEADSIZE']

            end_pos = chosen_genes.loc[i, 'start'] if direction < 0 else chosen_genes.loc[i, 'end']
            end_pos = end_pos//args['BEADSIZE'] + direction*((end_pos % args['BEADSIZE']) < (args['BEADSIZE']//2))
            end_pos -= chosen_slice[1]//args['BEADSIZE']

            load_prob = args['TIMESTEP'] / (chosen_genes.loc[i, 'rnap_sep'] / args['RNAP_VEL'])

            intron_list = full_df[(full_df['transcript_id'] == chosen_genes.loc[i, 'transcript_id']) & (full_df['type'] == 'intron')].copy()
            intron_list['len'] = intron_list['end'] - intron_list['start']
            intron_list = intron_list[intron_list['len'] > 0][['start', 'end', 'len']].values
            introns = []
            prev_cut_len = 0
            for *st_en, l in intron_list[::direction]:
                cut_pos = abs(st_en[(direction+1)//2] - chosen_genes.loc[i, ['start', 'end'][(direction-1)//2]]) - prev_cut_len
                prev_cut_len += l
                introns += [(cut_pos, l)]

            chosen_genes_beads += [[start_pos, direction, end_pos, load_prob, introns]]
            chosen_genes_beads += [[N - start_pos - 1, -1*direction, N - end_pos - 1, load_prob, introns]] # Make sister chromatid gene
    return chosen_genes_beads

def make_one_step(
    rnaps = None,
    occupied_rnap = None,
    chosen_genes_beads = None,
    condensins = None,
    occupied = None,
    maxcondesins = None,
    cohesins = None,
):
    global args
    
    if (rnaps is not None) and (occupied_rnap is not None) and (chosen_genes_beads is not None):
        translocate_rnap(rnaps, occupied_rnap, chosen_genes_beads, args, occupied = occupied, cohesins = cohesins, condensins = condensins)
    if (condensins is not None) and (occupied is not None) and (maxcondesins is not None):
        translocate_condensin(condensins, occupied, args, maxcondesins, cohesins = cohesins, occupied_rnap = occupied_rnap)
    if (cohesins is not None) and (occupied is not None):
        translocate_cohesin(cohesins, occupied, args, occupied_rnap = occupied_rnap)
    
def run_one_simulation(rnaps, occupied_rnap, occupied, cohesins, condensins):
    global N, args, steps, chosen_genes_beads, steps_wo_saving, max_rnaps, maxcondesins, N_coh

    cur_rnap_pos = []
    cur_brists = []
    cur_cond_pos = []
    cur_coh_pos = []
    for i in range(steps):
        for _ in range(steps_wo_saving):
            make_one_step(rnaps, occupied_rnap, chosen_genes_beads, condensins, occupied, maxcondesins, cohesins)
            
        rnap_positions = [rnap.position for rnap in rnaps] + [-1 for _ in range(max_rnaps - len(rnaps))]
        bristles = [rnap.bristle for rnap in rnaps] + [-1 for _ in range(max_rnaps - len(rnaps))]
        cond_positions = [(cond.left.pos, cond.right.pos) for cond in condensins] + [(-1,-1)]*(maxcondesins - len(condensins))
        coh_positions = [(coh.left.pos, coh.right.pos) for coh in cohesins] + [(-1,-1)]*(N_coh - len(cohesins))

        cur_rnap_pos.append(rnap_positions)  # appending current positions to an array
        cur_brists.append(bristles)
        cur_cond_pos.append(cond_positions)
        cur_coh_pos.append(coh_positions)

    return cur_rnap_pos, cur_brists, cur_cond_pos, cur_coh_pos

def params_generator(params, n):
    for _ in range(n):
        yield params

        
        
        
### Draw arcs

def plot_interaction(
    l, r, n_lef=0,
    height_factor=1.0,
    lw_factor=1.0,
    lw_max=5.0,
    max_height = 150,
    height=None,
    y_base=30,
    plot_text=True,
    orientation = 1,
    color='tomato',
    **kwargs,
):
    """Visualize an individual loop with an arc diagram.
    """
    arc_center = ((l+r)/2, y_base)
    arc_height = (min(max_height, (r-l)/2.0*height_factor * orientation)
                  if (height is None) else height)

    arc = Arc(xy=arc_center,
            width=r-l,
            height=arc_height,
            theta1 = min(0, 180 * orientation),
            theta2 = max(0, 180 * orientation),
            alpha = 0.45,
            lw=min(lw_max, 1.0*n_lef*lw_factor),
            color=color,
            capstyle='round')
    plt.gca().add_artist(arc)

    if n_lef > 1 and arc_center[0] < plt.xlim()[1] and plot_text:
        plt.text(x=arc_center[0],
                 y=arc_center[1]+arc_height/2+30,
                 horizontalalignment='center',
                 verticalalignment='center',
                 s=str(int(n_lef)),
                 fontsize=20,
                 color=color
                )
        
        
def plot_cohesin(
    l, r, n_lef=0,
    y_base = [0, 0],
    lw_factor=1.0,
    lw_max=5.0,
    color='seagreen',
    **kwargs, 
):
    """Visualize an individual cohesin with a line diagram.
    """
    poly = Polygon(
        xy = [[x0,x1] for x0,x1 in zip([l, r], [y_base[0], y_base[1]])],
        closed=False,
        lw=min(lw_max, 1.0*n_lef*lw_factor),
        color=color,
        alpha = 0.45,
    )
    plt.gca().add_artist(poly)
    

def prepare_canvas(
    L,
    site_width_bp = None,
    single_strand = True,
    max_y = 200,
    ):
#     plt.figure(figsize=(8,8))
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['bottom'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().tick_params(
        which='both',
        bottom='on',
        right='off',
        left='off',
        direction='out',
        top='off')
    plt.yticks([])
    
    if not single_strand:
        assert L%2 == 0, f'Sites number must be even if two_strands is True (now {L} sites)'
        plt.xlim(-1, L//2+1)
        plt.ylim(-max_y*1.2, max_y*1.2)
        plt.axhline(int(max_y*0.1), color='gray', lw=2, zorder=-1)
        plt.axhline(int(-max_y*0.1), color='gray', lw=2, zorder=-1)
    else:
        plt.xlim(-1, L+1)
        plt.ylim(0, max_y)
        plt.axhline(int(max_y*0.1), color='gray', lw=2, zorder=-1)

    if site_width_bp:
        plt.xticks([100*i*1000/float(site_width_bp) for i in range(16)],
                   [100*i for i in range(16)],
                   fontsize=20)
        plt.xlabel('chromosomal position, kb', fontsize=20)
        
    
    
def plot_lefs_single_strand(
        l_sites, 
        r_sites,
        y = [20],
        colors='tomato',
        **kwargs):
    """Plot an arc diagram for a list of loops.
    """
    
    unique_connections = pd.DataFrame(np.vstack([l_sites,r_sites]).T, columns=['l_sites', 'r_sites']).value_counts().reset_index()
    for i, (ls, rs, nl) in enumerate(unique_connections.values):
        plot_interaction(
            ls,
            rs,
            nl,
            y_base = y[0],
            color=colors[i] if (type(colors) in (list, tuple, np.ndarray)) else colors,
            **kwargs) 
        
def plot_lefs_sister_strands(
        l_sites, 
        r_sites, 
        L = 0,
        y = [20, -20],
        colors='tomato',
        **kwargs):
    """Plot an arc diagram for a list of loops.
    """
    
    unique_connections = pd.DataFrame(np.vstack([l_sites,r_sites]).T, columns=['l_sites', 'r_sites']).value_counts().reset_index()
    
    for i, (ls, rs, nl) in enumerate(unique_connections.values):
        l_orient = np.sign(ls-L//2)
        r_orient = np.sign(rs-L//2)
        if l_orient == r_orient:
            plot_interaction(
                L//2 - abs(ls-L//2),
                L//2 - abs(rs-L//2),
                nl,
#                 color=colors[i] if (type(colors) in (list, tuple, np.ndarray)) else colors,
                y_base = y[int(l_orient+1)//2],
                orientation = -l_orient,
                **kwargs)
        else:
            plot_cohesin(
                L//2 - abs(ls-L//2),
                L//2 - abs(rs-L//2),
                nl,
                y_base = y,
#                 color=colors[i] if (type(colors) in (list, tuple, np.ndarray)) else colors,
                **kwargs,
            )
            
            
def plot_rnap(
    site, bristle, n_brist=0,
    height_factor=1.0,
    lw_factor=1.0,
    lw_max=5.0,
    max_height = 150,
    height=None,
    y_base=30,
    plot_text=True,
    orientation = 1,
    color='#000080',
    **kwargs,
):
    circle = Circle(
        (site, y_base), 
        radius = 1,
        color = color,
        alpha = 0.35,
    )
    plt.gca().add_artist(circle)
    
    if np.sign(bristle) > 0:
        y_fin = min(y_base + bristle*height_factor, y_base + max_height)
    else:
        y_fin = max(y_base - abs(bristle)*height_factor, y_base - max_height)
        
    poly = Polygon(
        xy = [[x0,x1] for x0,x1 in zip([site, site], [y_base, y_fin])],
        closed=False,
        lw = min(lw_max, 1.0*n_brist*lw_factor),
        color=color,
        alpha = 0.35,
    )
    plt.gca().add_artist(poly)

def max_brist_and_counts(df):
    return pd.Series({'bristles': df['bristles'].max(), 'counts': len(df)})
    
def plot_rnaps_single_strand(
    sites, 
    bristles,
    y = [20],
    colors='#000080',
    **kwargs
):
    rnaps_df = pd.DataFrame(np.vstack([sites, bristles]).T, columns=['sites', 'bristles']).groupby('sites').apply(max_brist_and_counts).reset_index()
    for i, (site, bristle, nl) in enumerate(rnaps_df.values):
        plot_rnap(
            site,
            bristle,
            nl,
            y_base = y[0],
            color=colors[i] if (type(colors) in (list, tuple, np.ndarray)) else colors,
            **kwargs)
        
def plot_rnaps_sister_strands(
        sites, 
        bristles, 
        L = 0,
        y = [20, -20],
        colors='#000080',
        **kwargs
):
    rnaps_df = pd.DataFrame(np.vstack([sites, bristles]).T, columns=['sites', 'bristles']).groupby('sites').apply(max_brist_and_counts).reset_index()
    for i, (site, bristle, nl) in enumerate(rnaps_df.values):
        if site < L//2:
            plot_rnap(
                site,
                bristle,
                nl,
                y_base = y[0],
                color=colors[i] if (type(colors) in (list, tuple, np.ndarray)) else colors,
                **kwargs)
        else:
            plot_rnap(
                L//2 - abs(site-L//2),
                -bristle,
                nl,
                y_base = y[1],
                color=colors[i] if (type(colors) in (list, tuple, np.ndarray)) else colors,
                **kwargs)
        

def draw_gif_by_coords(
    out_name,
    path_to_rnap_h5 = None,
    path_to_coh_h5 = None,
    path_to_cond_h5 = None,
    first_frame = 0,
    n_frames = None,
    save_each_frame = 1,
    time_interval = 30, 
    single_strand = True, 
    shift_k = 0.1,
    rnap_heigh_factor = 3.0,
    arc_heigh_factor = 6.0,
    base_resolution = 1,
    out_resolution = int(1e3),
    **kwargs,
):
    
    fig, ax = plt.subplots()
    
    L = 0
    
    if path_to_rnap_h5 is not None:
        rnap_h5 = h5py.File(path_to_rnap_h5, mode='r')
        L = rnap_h5.attrs['N']
        max_brist_len = rnap_h5['bristles'][:].max()
    if path_to_coh_h5 is not None:
        coh_h5 = h5py.File(path_to_coh_h5, mode='r')
        L = coh_h5.attrs['N']
    if path_to_cond_h5 is not None:
        cond_h5 = h5py.File(path_to_cond_h5, mode='r')
        L = cond_h5.attrs['N']
    
    L = L *base_resolution//out_resolution
    
    def draw_frame(frame):
        plt.gca().clear()
        prepare_canvas(L, single_strand = single_strand, max_y = L)
        
        if path_to_rnap_h5 is not None:
            positions = rnap_h5['positions'][frame,:]
            bristles = rnap_h5['bristles'][frame,:]
            rnap_sites = positions[positions > 0] *base_resolution//out_resolution
            rnap_bristles = bristles[bristles > 0] *base_resolution//out_resolution

            if len(rnap_sites) > 0:
                if single_strand:
                    plot_rnaps_single_strand(
                        sites = rnap_sites, 
                        bristles = rnap_bristles,
                        y = [L*shift_k],
                        max_height = L,
                        plot_text = False,
                        height_factor = rnap_heigh_factor,
                        lw_factor = 1.0,
                        lw_max = 5.0,
                    )
                else:
                    plot_rnaps_sister_strands(
                        sites = rnap_sites, 
                        bristles = rnap_bristles,
                        L = L,
                        y = [(L/2)*shift_k, -(L/2)*shift_k],
                        max_height = L,
                        plot_text = False,
                        height_factor = rnap_heigh_factor,
                        lw_factor = 1.0,
                        lw_max = 5.0,
                    )
                    
            if path_to_coh_h5 is not None:
                l_sites = coh_h5['positions'][frame,:,0]
                r_sites = coh_h5['positions'][frame,:,1]
                l_sites = l_sites[l_sites > 0] *base_resolution//out_resolution
                r_sites = r_sites[r_sites > 0] *base_resolution//out_resolution
                assert l_sites.shape == r_sites.shape
                
                if len(l_sites) > 0:
                    if single_strand:
                        plot_lefs_single_strand(
                            l_sites = l_sites, 
                            r_sites = r_sites,
                            y = [L*shift_k],
                            max_height=L,
                            plot_text=False,
                            height_factor=arc_heigh_factor,
                            lw_factor=1.0,
                            lw_max=5.0,
                        )
                    else:
                        plot_lefs_sister_strands(
                            l_sites=l_sites, 
                            r_sites=r_sites,
                            L = L,
                            y = [L/2*shift_k, -L/2*shift_k],
                            max_height=L,
                            plot_text=False,
                            height_factor=arc_heigh_factor,
                            lw_factor=1.0,
                            lw_max=5.0,
                        )
                
            if path_to_cond_h5 is not None:
                l_sites = cond_h5['positions'][frame,:,0]
                r_sites = cond_h5['positions'][frame,:,1]
                l_sites = l_sites[l_sites > 0] *base_resolution//out_resolution
                r_sites = r_sites[r_sites > 0] *base_resolution//out_resolution
                assert l_sites.shape == r_sites.shape
                
                if len(l_sites) > 0:
                    if single_strand:
                        plot_lefs_single_strand(
                            l_sites = l_sites, 
                            r_sites = r_sites,
                            y = [L*shift_k],
                            max_height=L,
                            plot_text=False,
                            height_factor=arc_heigh_factor,
                            lw_factor=1.0,
                            lw_max=5.0,
                        )
                    else:
                        plot_lefs_sister_strands(
                            l_sites=l_sites, 
                            r_sites=r_sites,
                            L = L,
                            y = [L/2*shift_k, -L/2*shift_k],
                            max_height=L,
                            plot_text=False,
                            height_factor=arc_heigh_factor,
                            lw_factor=1.0,
                            lw_max=5.0,
                        )

    ani = animation.FuncAnimation(fig = fig, 
                                  func = draw_frame, 
                                  frames = range(first_frame, first_frame + (n_frames or len(rnap_h5['positions'])), save_each_frame), 
                                  interval = time_interval,
                                 )
    ani.save(out_name)

def draw_interaction_heatmap(
    base_resolution,
    out_name = None,
    path_to_coh_h5 = None,
    path_to_cond_h5 = None,
    out_name = None,
    first_frame = 0,
    n_frames = None,
    save_each_frame = 1,
    out_resolution = int(1e3),
    interaction_presence_only = False,
    log_scale = False,
    figsize = (5,5),
    sector = None,
    chunk = 1000,
):
    fig, ax = plt.subplots(figsize = figsize)
    
    L = 0
    
    if path_to_coh_h5 is not None:
        coh_h5 = h5py.File(path_to_coh_h5, mode='r')
        L = coh_h5.attrs['N']
        n_frames = n_frames or len(coh_h5['positions'])
        
    if path_to_cond_h5 is not None:
        cond_h5 = h5py.File(path_to_cond_h5, mode='r')
        L = cond_h5.attrs['N']
        n_frames = n_frames or len(cond_h5['positions'])
    
    rows = []
    cols = []
    bins = np.linspace(first_frame, first_frame + n_frames, max(20//int(1e0)+1, 2), dtype=int) # chunks boundaries
    for st,end in zip(bins[:-1], bins[1:]):
        save_each_frame
        if path_to_coh_h5 is not None:
            mask = (coh_h5['positions'][st:end:save_each_frame,:,0] >= 0)
            rows += coh_h5['positions'][st:end:save_each_frame][mask,0].reshape(-1).tolist() + (L - 1 - coh_h5['positions'][st:end:save_each_frame][mask,1]).reshape(-1).tolist()
            cols += (L - 1 - coh_h5['positions'][st:end:save_each_frame][mask,1]).reshape(-1).tolist() + coh_h5['positions'][st:end:save_each_frame][mask,0].reshape(-1).tolist()
        assert max(rows) < L/2
        assert max(cols) < L/2
            
        if path_to_cond_h5 is not None:
            mask_f = (cond_h5['positions'][st:end:save_each_frame,:,0] < L/2) & (cond_h5['positions'][st:end:save_each_frame,:,0] >= 0)
            mask_s = (cond_h5['positions'][st:end:save_each_frame,:,0] >= L/2) & (cond_h5['positions'][st:end:save_each_frame,:,0] >= 0)
            rows += cond_h5['positions'][st:end:save_each_frame][mask_f,0].reshape(-1).tolist() + (L - 1 - cond_h5['positions'][st:end:save_each_frame][mask_s,0]).reshape(-1).tolist() + \
                    cond_h5['positions'][st:end:save_each_frame][mask_f,1].reshape(-1).tolist() + (L - 1 - cond_h5['positions'][st:end:save_each_frame][mask_s,1]).reshape(-1).tolist()
            cols += cond_h5['positions'][st:end:save_each_frame][mask_f,1].reshape(-1).tolist() + (L - 1 - cond_h5['positions'][st:end:save_each_frame][mask_s,1]).reshape(-1).tolist() + \
                    cond_h5['positions'][st:end:save_each_frame][mask_f,0].reshape(-1).tolist() + (L - 1 - cond_h5['positions'][st:end:save_each_frame][mask_s,0]).reshape(-1).tolist()
        assert max(rows) < L/2
        assert max(cols) < L/2
    
    rows = np.array(rows) * base_resolution
    cols = np.array(cols) * base_resolution
    
    if sector is None:
        rows = np.round(rows / out_resolution).astype(int)
        cols = np.round(cols / out_resolution).astype(int)
        shape = (np.round(L/2 * base_resolution / out_resolution).astype(int)+1, 
                 np.round(L/2 * base_resolution / out_resolution).astype(int)+1)  
    else:
        assert len(sector) == 2
        assert (len(sector[0]) == len(sector[1])) and (len(sector[0]) == 2)
        mask = (rows >= sector[1][0]) & (rows < sector[1][1]) & (cols >= sector[0][0]) & (cols < sector[0][1])
        rows = rows[mask]
        cols = cols[mask]
        shape = (np.round((sector[1][1] - sector[1][0]) / out_resolution).astype(int)+1, 
                 np.round((sector[0][1] - sector[0][0]) / out_resolution).astype(int)+1)
        
        rows = np.round((rows - sector[1][0]) / out_resolution).astype(int)
        cols = np.round((cols - sector[0][0]) / out_resolution).astype(int)
    
    assert len(rows) == len(cols)    
    interaction_matrix = sp.coo_matrix((np.ones_like(rows), (rows, cols)), shape = shape).toarray()

    if interaction_presence_only: 
        interaction_matrix = interaction_matrix > 0
    elif log_scale: 
        interaction_matrix = np.nan_to_num(np.log10(interaction_matrix+1), nan=0.0, posinf=0.0, neginf=0.0)
                     
    sns.heatmap(interaction_matrix, square = True)
    
    if out_name is not None:
        plt.savefig(out_name)

### Other
def gtf_reader(file_path, transcripts_only = False, stop_after = None):
    genes_list = []
    with open(file_path) as gtf_file:
        line_len = []
        black_list_genes = []
        all_lines = gtf_file.readlines()
        for line_n,line in enumerate(all_lines):
            if line_n == stop_after:
                break
            if '#' not in line:
                line = line.split('\n')[0]
                line = re.split('[;\t]', line)
                
                if (transcripts_only and line[2] == 'transcript') or not (transcripts_only):
                    gene_id = line[8].split(' ')[-1][1:-1]

                    black_list = False
    #                 black_list = sum([
    #                     y[0] <= int(line[3]) <= y[1] or \
    #                     y[0] <= int(line[4]) <= y[1] \
    #                     for y in tel_cen
    #                 ])
                    if black_list:
                        black_list_genes += [gene_id]
                    if not gene_id in black_list_genes:
                        if 'TPM' in line[-2]:
                            TPM = float(line[-2].split(' ')[-1][1:-1])
                        if 'FPKM' in line[-3]:
                            FPKM = float(line[-3].split(' ')[-1][1:-1])
                        if 'cov' in line[-4]:
                            cov = float(line[-4].split(' ')[-1][1:-1])
                        else:
                            cov = float(line[-2].split(' ')[-1][1:-1])
                        tmp = {
                            'type': line[2],
                            'chrom': line[0],
                            'start': int(line[3]),
                            'end': int(line[4]),
                            'direction': line[6],
                            'gene_id': line[8].split(' ')[-1][1:-1],
                            'transcript_id': line[9].split(' ')[-1][1:-1],
                            'TPM': TPM,
                            'FPKM': FPKM,
                            'cov': cov,
                        }
                        genes_list += [tmp]
            if line_n % 1000 == 0: print(f'{int(line_n/len(all_lines)*100)}% ({line_n} / {len(all_lines)}) are read' + ' '*200, end = '\r')
    return pd.DataFrame(genes_list, columns = ['type', 'chrom', 'start', 'end', 'direction', 'gene_id', 'transcript_id', 'TPM', 'FPKM', 'cov'])


def verbose_timedelta(delta):
    d, s = divmod(delta, 60*60*24)
    h, s = divmod(s, 60*60)
    m, s = divmod(s, 60)
    s, ms = divmod(s, 1)
    ms = ms*1000
    labels = ['d', 'h', 'm', 's', 'ms']   
    dhms = ['%s%s' % (i, lbl) for i, lbl in zip([int(d), int(h), int(m), int(s), int(ms)], labels)]
    for start in range(len(dhms)):
        if not dhms[start].startswith('0'):
            break
    for end in range(len(dhms)-1, -1, -1):
        if not dhms[end].startswith('0'):
            break  
    return ':'.join(dhms[start:end+1])